/*
 * This module implements a simple character device.
 *
 * Create the device file using mknod:
 *
 * $ sudo rm -f /dev/my-cdev
 * $ sudo insmod ./cdev.ko
 * $ num=$(awk '/my-cdev/ { print $1 }' /proc/devices)
 * $ sudo mknod /dev/my-cdev c $num 0
 *
 * Check that the mutex does do its job:
 *
 * $ sleep 10 < /dev/my-cdev &
 * $ cat /dev/my-cdev
 *
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/cdev.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/uaccess.h>
#include <linux/mutex.h>

MODULE_AUTHOR("Juergen Schoenwaelder");
MODULE_LICENSE("Dual BSD/GPL");
MODULE_DESCRIPTION("Simple character device demonstration module.");

#define kwarn(FMT, ...) \
        printk(KERN_WARNING "[%s] "FMT"\n", __this_module.name, ##__VA_ARGS__)

#define kinfo(FMT, ...) \
        printk(KERN_INFO "[%s] "FMT"\n", __this_module.name, ##__VA_ARGS__)

static const char *dev_name = "my-cdev";
static dev_t mydev;
static char buffer[64];

static char *data = NULL;
static size_t size = 0;
static size_t used = 0;

static DEFINE_MUTEX(my_mutex);

static ssize_t
my_read(struct file *filp, char __user *buf, size_t count, loff_t *f_pos)
{
    if (*f_pos >= used) {
	return 0;
    }

    /* limit count to the amount of data left (if necessary) */

    if (used - *f_pos < count) {
	count = used - *f_pos;
    }

    /* copy count characters starting at f_pos to user space */
    
    if (copy_to_user(buf, data + *f_pos, count) > 0) {
	return -EFAULT;
    }

    /* update file offset and return that we copied count characters */
    
    *f_pos += count;
    return count;
}

static ssize_t
my_write(struct file *file, const char __user *buf, size_t count, loff_t *f_pos)
{
    if (count > sizeof(buffer)) {
	count = sizeof(buffer);
    }

    /* copy count characters from user space */
    
    if (copy_from_user(buffer, buf, count)) {
	return -EFAULT;
    }

    /* reallocate buffer space if needed */

    if (*f_pos + count > size) {
	size = *f_pos + count;
	data = krealloc(data, size, GFP_KERNEL);
	if (! data) {
	    return -ENOMEM;
	}
    }

    /* copy the data where it belongs, update how much data is used */
    
    memcpy(data + *f_pos, buffer, count);
    if (*f_pos + count > used) {
	used = *f_pos + count;
    }

    /* update file offset and return that we copied count characters */
    
    *f_pos += count;
    return count;
}

static int
my_open(struct inode *inode, struct file *file)
{
    if (!mutex_trylock(&my_mutex)) {
	return -EBUSY;
    }
    
    if (file->f_mode & FMODE_WRITE) {
	if (data) {
	    kfree(data);
	    data = NULL;
	    size = 0;
	    used = 0;
	}
    }
    return 0;
}

static int
my_release(struct inode *inode, struct file *file)
{
    mutex_unlock(&my_mutex);
    return 0;
}

static struct cdev my_cdev;

static struct file_operations my_fops = {
    .owner = THIS_MODULE,
    .open = my_open,
    .read = my_read,
    .write = my_write,
    .release = my_release,
};

static int __init
my_cdev_init(void)
{
    int rc;
    
    rc = alloc_chrdev_region(&mydev, 0, 1, dev_name);
    if (rc < 0) {
	kwarn("failed to allocate cdev major/minor device numbers");
	return rc;
    }
    kinfo("device \"%s\" has device number %s",
	  dev_name, format_dev_t(buffer, mydev));

    cdev_init(&my_cdev, &my_fops);
    my_cdev.owner = THIS_MODULE;
    rc = cdev_add(&my_cdev, mydev, 1);
    if (rc < 0) {
	unregister_chrdev_region(mydev, 1);
	return rc;
    }

    mutex_init(&my_mutex);

    return 0;
}

static void __exit
my_cdev_exit(void)
{
    if (data) {
	kfree(data);
    }

    mutex_destroy(&my_mutex);
    
    cdev_del(&my_cdev);
    unregister_chrdev_region(mydev, 1);
}

module_init(my_cdev_init);
module_exit(my_cdev_exit);
